############################################
# Classes for migrating SQLite3 database. 
# 
# Author: Ondrej Fabianek
#
# Description:
#
#  Only 'DBMigrator' should be imported to other files. The
#  rest of the classes serve only as internal implementation 
#  of the main class.
#
#  It uses output of ".dump" sqlite command to build a model of 
#  the  database in its internal structures. This model can then
#  be modified and later be stored back to the database file by 
#  calling commit() method.
#  

#############################################
# Imports from an absolute path. 'name' is
# for example name of a class or a function.
def import_from(src_file_path, name):
  import importlib.util
  spec = importlib.util.spec_from_file_location(name, src_file_path)
  foo = importlib.util.module_from_spec(spec)
  spec.loader.exec_module(foo)
  return foo
  
import subprocess, copy, os

try:
  from database import Database
except:
  # ##################### !!! ####################### #
  # TODO: Edit 'dir_vpnportal' path if you move this 
  # script to another directory!
  dir_vpnportal = os.path.dirname(os.path.realpath(__file__)) + "/../.."    
  file_functions = dir_vpnportal+"/install/scripts/database.py"
  import_from(file_functions, "Database")
    

class DB_Index:
  #############################################
  # Example of index: 
  # CREATE UNIQUE INDEX UNIQ_88BDF3E9A0D96FBF ON app_user (email_canonical);
  def __init__(self, db_index):
    self.name       = self._parse_name(db_index)
    self.table_name = self._parse_table_name(db_index)
    self.col_name   = self._parse_col_name(db_index)
  
  def _parse_name(self, db_index):
    return db_index.split()[3]
    
  def _parse_table_name(self, db_index):
    return db_index.split()[5]
    
  def _parse_col_name(self, db_index):
    return db_index.split()[6][1:-1] # Strip out the brackets
    
  def to_str(self):
    return 'CREATE UNIQUE INDEX '+self.name+' ON '+self.table_name+' ('+self.col_name+');\n'
  
      
class DB_Trigger:
  #############################################
  # Example of trigger: 
  # CREATE TRIGGER decrement_pending_count1
  # AFTER DELETE ON pending_lan_table
  # BEGIN
  #  UPDATE device_table SET pending_count = pending_count - 1 WHERE device_id = OLD.device_id;
  # END;
  def __init__(self, trigger):
    self.name = self._parse_name(trigger)
    self.body = self._parse_body(trigger)
  
  def to_str(self):
    if "END;" not in self.body:
      return 'CREATE TRIGGER ' + self.name + self.body + ";\nEND;\n"     
    return 'CREATE TRIGGER ' + self.name + self.body
    
  def replace_str_in_body(self, old_str, new_str):
    self.body = self.body.replace(old_str, new_str)
  
  def _parse_name(self, trigger):
    start = trigger.find("TRIGGER")+8
    end   = trigger.find("\n")
    return trigger[start:end]
  
  def _parse_body(self, trigger):
    start = trigger.find("\n")
    return trigger[start:]

class DB_Restri:
  #############################################
  # Example of restri: "PRIMARY KEY (name, date)"
  def __init__(self, restri):
    start = restri.find("(")+1
    end   = restri.rfind(")")
    self.col_names = restri[start:end].split(",")
    self.key_word  = restri[:start-1]
    
  def to_str(self):
    result = " " + self.key_word + " ("
    for name in self.col_names:
      result += name + ","
    result = result[:-1]
    result += "),\n"
    return result 

class DB_Entry:
  #############################################
  # Examples of 'insert': 'INSERT INTO device_table (device_id, serial) VALUES(5, 'adv-00.0A.14.DE.F0.A1');'
  #                       'INSERT INTO device_table VALUES(5, 'adv-00.0A.14.DE.F0.A1');'
  def __init__(self, insert, col_names=[]):
    self.values     = self._extract_values(insert)
    self.table_name = self._extract_table_name(insert)
    self.col_names  = self._extract_col_names(insert)

  def update_value(self, index, value):
    self.values[index] = value
  
  def erase_other_than(self, col_names):
    to_erase = []
    if not self.col_names:
      RuntimeError("This method requires column names to be set for the entry.")
    for name in self.col_names:
      if name not in col_names:
        to_erase.append(name)
    for name in to_erase:
      index = self._get_col_index(name)
      self.col_names.pop(index)
      self.values.pop(index)      
  
  def get_value_by_index(self, index):
    return self.values[index]
  
  def set_col_names(self, names):
    if len(names) != len(self.values):
      raise RuntimeError("Invalid parameter: number of columns must match number of values.")
    self.col_names = names
    
  def del_value_by_col(self, col_name):
    index = 0
    for n in self.col_names:
      if n == col_name:
        self.del_value(index)
        return
      index += 1
    raise RuntimeError("Could not find column " + col_name + ".")  
    
  def del_value(self, index):
    self.values.pop(index) # pop() removes value on the given index
    if self.col_names:
      self.col_names.pop(index)
  
  def add_value(self, value, col_name="", pos=0):
    if not value:
      value = "\'\'"
    self.values.insert(pos, value)
    if col_name:
      self.col_names.insert(pos, col_name)
    if not col_name and self.col_names:
      raise RuntimeError("Invalid parameters: Expecting to receive a column name.")
          
  def to_str(self):
    result = 'INSERT INTO ' + self.table_name
    if self.col_names:
      result += "(" + (",".join(self.col_names)) + ')'
    result += ' VALUES('
    for v in self.values:
      result += v.replace("\\'", "'") + ","
    result = result[:-1]
    result += ");\n"
    return result

  def _get_col_index(self, col_name):
    index = 0
    for name in self.col_names:
      if name == col_name:
        return index
      index += 1
    raise RuntimeError("No value found for column " + col_name + ".")
    
  def _get_value_by_col(self, col_name):
    index = 0
    for c in self.col_names:
      if c == col_name:
        return self.values[index]
      index += 1
    raise RuntimeError("Column " + col_name + " was not found.")
  
  def _extract_next_value(self, insert):
    if insert.find(",") < 0: # If we are at the end of the command
      value = insert
      return value, None
    elif insert[0] == "'": # A string value
      end   = insert.find("',")+1
      value = insert[:end]
      return value, insert[end+1:]
    elif insert[0] != "r": # An integer value
      end = insert.find(",")
      value = insert[:end]
      return value, insert[end+1:]
    elif insert[:8] == "replace(": # A "replace()" function added by sqlite to make our life difficult.
      insert = insert[8:] # Get rid of "replace("
      end   = insert[1:].find("'")+2
      value = insert[:end]
      repend = insert[end:].find(")),")+3+end
      return value, insert[repend:]
    else:
      raise RuntimeError("Could not parse values of: " + insert)
    
  def _extract_values(self, insert):
    start  = insert.find("VALUES(")+7
    end    = insert[start:].rfind(")")+start
    insert = insert[start:end].replace("\\n", "\n").replace("\\\n", "\n")
    value, insert = self._extract_next_value(insert)
    result = []
    while insert != None:
      result.append(value)
      value, insert = self._extract_next_value(insert)
    result.append(value)
    return result
  
  def _extract_table_name(self, insert):
    start = insert.find("VALUES(")+7
    return insert[:start].split()[2]
  
  def _extract_col_names(self, insert):
    stop  = insert.find("VALUES")
    start = insert[:stop].find("(")
    if start < 0:
      return []
    end   = insert[start:stop].find(")")
    return insert[start:end].split(",") 
    
class DB_Column:
  #############################################
  # Example of 'column_def':  'online integer NOT NULL DEFAULT 0'
  def __init__(self, column_def):
    c = column_def.replace(",", "", 1).split()
    self.name = c[0]
    self.details = c[1:]
  
  def get_name(self):
    return self.name
  
  def get_type(self):
    return self.details[0]  
    
  def edit(self, new_name, details=""):
    self.name = new_name
    if details:
      self.details = details.split()
      
  def to_str(self):
    result = "   " + self.name + "\t"
    for d in self.details:
      result += d + " "
    result = result[:-1]
    if '--(DC2Type:array)' in result:
      result += "\n,\n"
    else:
      result += ",\n"
    return result
 
class DB_Table:
  #############################################
  #
  def __init__(self, schema, inserts=[]):
    self.name    = self._extract_name(schema)
    self.columns = self._extract_columns(schema)
    self.restri  = self._extract_restrictions(schema)
    self.entries = self._extract_entries(inserts)

  def set_name(self, new_name):
    self.name = new_name
    for e in self.entries:
      e.table_name = new_name
  
  def update_col_values(self, col_name, values, from_back=False):
    i = self._get_col_index(col_name)
    entries = reversed(self.entries) if from_back else self.entries 
    vals    = reversed(values) if from_back else values 
    for e, v in zip(entries, vals):
      e.update_value(i, v)
  
  #############################################
  # Functor takes 1 argument (current_value of  
  # the column) and returns a new value.
  def update_col_values_f(self, col_name, functor):
    values_old = self.get_col_values(col_name)
    values_new = []
    for old_v in values_old:
      values_new.append(functor(old_v))
    self.update_col_values(col_name, values_new)
  
  def count_entries(self):
    return len(self.entries)
  
  def get_col_values(self, col_name, count=999999, from_back=False):
    result      = []
    i           = self._get_col_index(col_name)
    all_entries = reversed(self.entries) if from_back else self.entries 
    loop = 0
    for e in all_entries:
      if loop == count:
        break
      loop += 1
      result.append(e.get_value_by_index(i))
    return result
    
  def get_col_names(self):
    result = []
    for c in self.columns:
      result.append(c.name)
    return result
  
  def make_copy(self, name_of_copy, col_names):
    nc = copy.deepcopy(self)
    nc.set_name(name_of_copy)
    for c in self.columns:
      if c.name not in col_names:
        nc.del_column(c.name)
    return nc
  
  def truncate(self):
    self.entries = []
  
  def edit_column(self, old_name, new_name, details=""):
    c = self._get_col(old_name)
    if not c:
      raise RuntimeError("Column " + old_name + " not found in " + self.name + ".")
    c.edit(new_name, details)
  
  def del_column(self, col_name):
    index = self._get_col_index(col_name)
    self.columns.pop(index)
    for entry in self.entries:
      entry.del_value(index)
  
  def add_column(self, col_name, details, value='', pos=0):
    new_col = DB_Column(col_name + " " + details)
    self.columns.insert(pos, new_col)
    self._add_value_to_entries(value, pos)
  
  def _add_value_to_entries(self, value, pos=0):
    for entry in self.entries:
      entry.add_value(value, '', pos)  
  
  def to_str(self):
    result = "CREATE TABLE " + self.name + "(\n"
    result += self._columns_to_str()
    result += self._restri_to_str()
    result += "\n);\n" 
    result += self._entries_to_str()
    return result
  
  def _get_col(self, name):
    for c in self.columns:
      if c.name == name:
        return c
  
  def _get_col_index(self, col_name):
    index = -1
    for c in self.columns:
      index += 1
      if c.name == col_name:
        return index
    raise RuntimeError("Column " + col_name + " not found in table " + self.name + ".") 
    
  def _entries_to_str(self):
    result = ""
    for entry in self.entries:
      result += entry.to_str()
    return result
    
  def _columns_to_str(self):   
    result = "" 
    for column in self.columns:
      result += column.to_str()
    # if there are no rows like: "PRIMARY KEY (name, date)"
    # then we remove the last ',\n'
    if not self.restri:
      result = result[:-2] # remove the last ',\n'    
    return result

  def _restri_to_str(self):
    result = ""
    for r in self.restri:
      result += r.to_str()
    # If there was at least one restriction
    if self.restri:
      result = result[:-2] # remove the last ',\n'   
    return result
    
  def _extract_restrictions(self, schema):
    start = schema.find("(")+1
    end   = schema.rfind(")")
    cols  = schema[start:end].split('\n')
    result = []
    for col in cols:
      if col and 'PRIMARY KEY (' in col:
        result.append(DB_Restri(col))
    return result    
  
  def _extract_columns(self, schema):
    start  = schema.find("(")+1
    end    = schema.rfind(")")
    cols_s = schema[start:end]
    sep    = self._get_col_separator(cols_s)
    cols   = cols_s.split(sep)
    result = []
    for col in cols:
      if col.lstrip() and col[0] != ',' and 'PRIMARY KEY (' not in col:
        result.append(DB_Column(col))
    return result
  
  def _extract_name(self, schema):
    header = schema.split('(', 1)[0] # take everything after the first occurence of '('
    return header.split()[-1]

  def _extract_entries(self, inserts):
    result = []
    for row in inserts:
      result.append(DB_Entry(row))
    return result

  #############################################
  # Some database schemas do not put each column
  # on separate row. This function returns the 
  # character that can be best used to split the
  # columns.
  def _get_col_separator(self, cols_schema):
    count_types = cols_schema.count(" integer ") + \
                  cols_schema.count(" text ") + \
                  cols_schema.count(" string ") + \
                  cols_schema.count(" int ") +\
                  cols_schema.count(" INTEGER ") + \
                  cols_schema.count(" TEXT ") +\
                  cols_schema.count(" STRING ") +\
                  cols_schema.count(" VARCHAR") +\
                  cols_schema.count(" BOOLEAN ") +\
                  cols_schema.count(" DATETIME ") +\
                  cols_schema.count(" BLOB ")
    if count_types > cols_schema.count("\n"):
      return ','
    return '\n'      
      
    
class DBMigrator:
  #############################################
  #
  def __init__(self, db_path):
    self.db = Database(db_path)
    self.reload(self.db.sqlite_dump())

  def update_col_values(self, table_name, col_name, values, from_back=False):
    table = self._get_table(table_name)
    table.update_col_values(col_name, values, from_back)
    self._add_m_table(table.name)
      
  #############################################
  # 'functor' is a callback function that takes 
  # 1 parameter (old column value) and returns 
  # a new column value.
  def update_col_values_f(self, table_name, col_name, functor):
    table = self._get_table(table_name)
    table.update_col_values_f(col_name, functor)
    self._add_m_table(table.name)
      
  #############################################
  # Parses a database dump or scheme.
  def reload(self, raw_db_text=None):
    if not raw_db_text:
      raw_db_text = self.db.sqlite_dump()
    dump_output      = raw_db_text.replace("\\t", "\t").replace("\\n", "\n").split(";\n")
    self.tables      = self._extract_tables(dump_output)
    self.triggers    = self._extract_triggers(dump_output)
    self.dbindexes   = self._extract_dbindexes(dump_output)
    self.m_tables    = set() # A set of names of those tables that were modified 
    self.m_triggers  = set()
    self.m_dbindexes = set()

  def edit_trigger_body(self, name, str_to_replace, new_str):
    self._get_trigger(name).replace_str_in_body(str_to_replace, new_str)
    self.m_triggers.add(name)
    
  def create_table(self, schema):
    self.tables.append(DB_Table(schema))
    self._add_m_table(self.tables[-1].name)
    
  def create_trigger(self, schema):
    self.triggers.append(DB_Trigger(schema))
    self.m_triggers.add(self.triggers[-1].name)
      
  def count_entries(self, table_name):
    return self._get_table(table_name).count_entries()
  
  #############################################
  # Takes all entries from column 'src_col' in 'src_table' and writes them
  # into the 'dst_col' of 'dst_table'. If 'dst_table' contains more entries
  # than 'src_table' then only a subset of entries in 'dst_table', starting 
  # from the back (aka with the last record of the table), will be modified.
  def copy_column_values(self, src_table, dst_table, src_col, dst_col, from_back=True):
    s_table = self._get_table(src_table)
    d_table = self._get_table(dst_table)
    count   = d_table.count_entries()
    values  = s_table.get_col_values(src_col, count, from_back)
    d_table.update_col_values(dst_col, reversed(values))
    self._add_m_table(dst_table)
  
  #############################################
  # 'howto' is a dictionary that might for example look like this:
  # {"online":"$online", "name":"$rw_name", "validated":"1"}
  # Where keys are columns of the destination table and values are 
  # what should be written there. If a value starts with '$', then it
  # represents a name of a column in the source table, whose values we 
  # want to copy.
  def copy_entries(self, src_table, dst_table, howto):
    index = self._get_table_index(src_table)
    src_t = self.tables[index]
    index = self._get_table_index(dst_table)
    dst_t = self.tables[index]
    for entry in src_t.entries:
      nc = copy.deepcopy(entry)
      nc.table_name = dst_table
      nc.set_col_names(src_t.get_col_names())
      for dst_col, v in howto.items():
        if v.startswith('$'):
          tmp = nc._get_value_by_col(v[1:])
          nc.del_value_by_col(v[1:])
          nc.add_value(tmp, dst_col)
        else:
          nc.add_value(v, dst_col)
      nc.erase_other_than(dst_t.get_col_names())
      dst_t.entries.append(nc)
    self._add_m_table(dst_table)
    
  def derive_new_table(self, old_table, new_table, columns):
    i = self._get_table_index(old_table)
    new = self.tables[i].make_copy(new_table, columns)
    self.tables.append(new)
    self._add_m_table(new.name)
    
  def edit_column(self, table_name, old_name, new_name, details=""):
    i = self._get_table_index(table_name)
    self.tables[i].edit_column(old_name, new_name, details)
    self._add_m_table(table_name)
  
  def del_column(self, table_name, col_name):
    i = self._get_table_index(table_name)
    self.tables[i].del_column(col_name)
    self._add_m_table(table_name)
  
  def add_column(self, table_name, col_name, details, value='', pos=0):
    i = self._get_table_index(table_name)
    self.tables[i].add_column(col_name, details, value, pos)   
    self._add_m_table(table_name)
  
  def truncate_table(self, table_name):
    i = self._get_table_index(table_name)
    self.tables[i].truncate()
    self._add_m_table(table_name) 
  
  #############################################
  # Propagates the changed database model into the database file.
  # It only touches those tables (and their records) that were in
  # any way modified.
  def commit(self):
    cmd = ""
    for table_name in self.m_tables:
      cmd += "DROP TABLE IF EXISTS " + table_name + ";\n"
    for trigger_name in self.m_triggers:
      cmd += "DROP TRIGGER IF EXISTS " + trigger_name + ";\n"
    for dbindex_name in self.m_dbindexes:
      cmd += "DROP INDEX IF EXISTS " + dbindex_name + ";\n"
    cmd += self.to_str(only_modified=True)
    self.db.run(cmd)

  #############################################
  # If 'only_modified' is set to False, then the whole database 
  # model is printed as known to this object. 
  # If it is set to True, then only those commands are printed 
  # that would be applied on the database file when 'commit()'
  # is called.   
  def to_str(self, only_modified=False):
    result = ""
    for table in self.tables:
      if not only_modified or table.name in self.m_tables:
        result += table.to_str()
    for trigger in self.triggers:
      if not only_modified or trigger.name in self.m_triggers:
        result += trigger.to_str()
    for dbindex in self.dbindexes:
      if not only_modified or dbindex.name in self.m_dbindexes:
        result += dbindex.to_str()      
    return str(result)
  
  #############################################
  # Adds a table to the list of tables that have
  # been modified in some way. Also marks all 
  # indexes related to this table as modified.
  def _add_m_table(self, table_name):
    self.m_tables.add(table_name)
    for dbindex in self.dbindexes:
      if dbindex.table_name == table_name:
        self.m_dbindexes.add(dbindex.name)
  
  def _get_trigger(self, trigger_name):
    for t in self.triggers:
      if t.name == trigger_name:
        return t
    raise RuntimeError("Trigger " + trigger_name + " not found.")
  
  def _get_table(self, table_name):
    i = self._get_table_index(table_name)
    return self.tables[i]
  
  def _get_table_index(self, table_name):
    index = 0
    for table in self.tables:
      if table.name == table_name:
        return index
      index += 1
    raise RuntimeError("Table " + table_name + " does not exist.")
  
  def _extract_dbindexes(self, dump_output):
    result = []
    for cmd in dump_output:
      if cmd.lstrip().startswith('CREATE UNIQUE INDEX'):
        result.append(DB_Index(cmd))
    return result    
  
  def _extract_triggers(self, dump_output):
    result = []
    for cmd in dump_output:
      if cmd.lstrip().startswith('CREATE TRIGGER'):
        result.append(DB_Trigger(cmd))
    return result
  
  def _extract_tables(self, dump_output):
    result = []
    inserts = []
    schema = None
    for cmd in dump_output:
      cmd = cmd.lstrip()
      if cmd.startswith("CREATE TABLE") or cmd.startswith("create table"):
        if schema:
          result.append(DB_Table(schema, inserts))
          inserts = []
        schema = cmd 
      elif cmd.startswith("INSERT") or cmd.startswith("insert"):
        inserts.append(cmd)
    if schema:
      result.append(DB_Table(schema, inserts))
    return result    


class DBSchemeReader(DBMigrator):
  #############################################
  #
  def __init__(self, scheme_path):
    with open(scheme_path, 'r') as file:
      raw_scheme = file.read()
      self.reload(raw_scheme) 
      
  #############################################
  # Returns array of pairs (col_name, col_type).
  def get_columns(self, table_name):
    table = self._get_table(table_name)
    result = []
    for col in table.columns:
      result.append(col)
    return result

  #############################################
  # Returns array with table names.    
  def get_table_names(self):
    result = []
    for table in  self.tables:
      result.append(table.name)
    return result  
  
